==> webkit2gtk_2.51.3.orig.tar.xz: webkitgtk-2.51.3/Source/WTF/wtf/WeakListHashSet.h <==
/*
 * Copyright (C) 2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <algorithm>
#include <wtf/ListHashSet.h>
#include <wtf/WeakPtr.h>

namespace WTF {

// Value will be deleted lazily upon rehash or amortized over time. For manual cleanup, call removeNullReferences().
template <typename T, typename WeakPtrImplType>
class WeakListHashSet final {
    WTF_DEPRECATED_MAKE_FAST_ALLOCATED(WeakListHashSet);
public:
    using KeyType = WeakPtr<T, WeakPtrImplType>;
    using SetType = ListHashSet<KeyType>;
    using AddResult = typename SetType::AddResult;

    template <typename ListHashSetType, typename IteratorType>
    class WeakListHashSetIteratorBase {
    public:
        using iterator_category = std::forward_iterator_tag;
        using value_type = T;
        using difference_type = ptrdiff_t;
        using pointer = T*;
        using reference = T&;

    protected:
        WeakListHashSetIteratorBase(ListHashSetType& set, IteratorType position)
            : m_set { set }
            , m_position { position }
            , m_beginPosition { set.m_set.begin() }
            , m_endPosition { set.m_set.end() }
        {
            skipEmptyBuckets();
        }

        WeakListHashSetIteratorBase(const ListHashSetType& set, IteratorType position)
            : m_set { set }
            , m_position { position }
            , m_beginPosition { set.m_set.begin() }
            , m_endPosition { set.m_set.end() }
        {
            skipEmptyBuckets();
        }

        ~WeakListHashSetIteratorBase() = default;

        T* get() { return m_position->get(); }
        const T* get() const { return get(); }
        T& operator*() { return *get(); }
        const T& operator*() const { return *get(); }
        T* operator->() { return get(); }
        const T* operator->() const { return get(); }

        void advance()
        {
            ASSERT(m_position != m_endPosition);
            ++m_position;
            skipEmptyBuckets();
            m_set.increaseOperationCountSinceLastCleanup();
        }

        void advanceBackwards()
        {
            ASSERT(m_position != m_beginPosition);
            --m_position;
            while (m_position != m_beginPosition && !get())
                --m_position;
            m_set.increaseOperationCountSinceLastCleanup();
        }

        void skipEmptyBuckets()
        {
            while (m_position != m_endPosition && !get())
                ++m_position;
        }

        const ListHashSetType& m_set;
        IteratorType m_position;
        IteratorType m_beginPosition;
        IteratorType m_endPosition;
    };

    class WeakListHashSetIterator : public WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::iterator> {
    public:
        using Base = WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::iterator>;

        bool operator==(const WeakListHashSetIterator& other) const { return Base::m_position == other.Base::m_position; }

        using Base::get;
        using Base::operator*;
        using Base::operator->;

        WeakListHashSetIterator& operator++()
        {
            Base::advance();
            return *this;
        }

        WeakListHashSetIterator& operator--()
        {
            Base::advanceBackwards();
            return *this;
        }

    private:
        WeakListHashSetIterator(WeakListHashSet& map, typename SetType::iterator position)
            : Base { map, position }
        { }

        template <typename, typename> friend class WeakListHashSet;
    };

    class WeakListHashSetConstIterator : public WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::const_iterator> {
    public:
        using Base = WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::const_iterator>;

        bool operator==(const WeakListHashSetConstIterator& other) const { return Base::m_position == other.Base::m_position; }

        using Base::get;
        using Base::operator*;
        using Base::operator->;

        WeakListHashSetConstIterator& operator++()
        {
            Base::advance();
            return *this;
        }

        WeakListHashSetConstIterator& operator--()
        {
            Base::advanceBackwards();
            return *this;
        }

    private:
        WeakListHashSetConstIterator(const WeakListHashSet& map, typename SetType::const_iterator position)
            : Base { map, position }
        { }

        template <typename, typename> friend class WeakListHashSet;
    };

    using iterator = WeakListHashSetIterator;
    using const_iterator = WeakListHashSetConstIterator;

    iterator begin() { return WeakListHashSetIterator(*this, m_set.begin()); }
    iterator end() { return WeakListHashSetIterator(*this, m_set.end()); }
    const_iterator begin() const { return WeakListHashSetConstIterator(*this, m_set.begin()); }
    const_iterator end() const { return WeakListHashSetConstIterator(*this, m_set.end()); }

    iterator find(const T* value)
    {
        increaseOperationCountSinceLastCleanup();
        return WeakListHashSetIterator(*this, m_set.find(value));
    }

    iterator find(const T& value)
    {
        return find(&value);
    }

    const_iterator find(const T* value) const
    {
        increaseOperationCountSinceLastCleanup();
        return WeakListHashSetConstIterator(*this, m_set.find(value));
    }

    const_iterator find(const T& value) const
    {
        return find(&value);
    }

    bool contains(const T* value) const
    {
        increaseOperationCountSinceLastCleanup();
        return m_set.contains(value);
    }

    bool contains(const T& value) const
    {
        return contains(&value);
    }

    AddResult add(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.add(value);
    }

    AddResult add(const T& value)
    {
        return add(&value);
    }

    AddResult appendOrMoveToLast(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.appendOrMoveToLast(value);
    }

    AddResult appendOrMoveToLast(const T& value)
    {
        return appendOrMoveToLast(&value);
    }

    bool moveToLastIfPresent(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.moveToLastIfPresent(value);
    }

    bool moveToLastIfPresent(const T& value)
    {
        return moveToLastIfPresent(&value);
    }

    AddResult prependOrMoveToFirst(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.prependOrMoveToFirst(value);
    }

    AddResult prependOrMoveToFirst(const T& value)
    {
        return prependOrMoveToFirst(&value);
    }

    AddResult insertBefore(const T* beforeValue, const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.insertBefore(beforeValue, value);
    }

    AddResult insertBefore(const T& beforeValue, const T& value)
    {
        return insertBefore(&beforeValue, &value);
    }

    AddResult insertBefore(iterator it, const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.insertBefore(it.m_position, value);
    }

    AddResult insertBefore(iterator it, const T& value)
    {
        return insertBefore(it, &value);
    }

    bool remove(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.remove(value);
    }

    bool remove(const T& value)
    {
        return remove(&value);
    }

    bool remove(iterator it)
    {
        if (it == end())
            return false;

        // WeakPtr<T> can become null while 'it' is live.
        if (!it.get())
            return false;

        auto result = m_set.remove(it.m_position);
        amortizedCleanupIfNeeded();
        return result;
    }

    void clear()
    {
        m_set.clear();
        cleanupHappened();
    }

    unsigned capacity() const { return m_set.capacity(); }

    bool isEmptyIgnoringNullReferences() const
    {
        if (m_set.isEmpty())
            return true;

        auto onlyContainsNullReferences = begin() == end();
        if (onlyContainsNullReferences) [[unlikely]]
            const_cast<WeakListHashSet&>(*this).clear();
        return onlyContainsNullReferences;
    }

    bool hasNullReferences() const
    {
        unsigned count = 0;
        auto result = std::ranges::any_of(m_set, [&](auto& iterator) {
            ++count;
            return !iterator.get();
        });
        if (result)
            increaseOperationCountSinceLastCleanup(count);
        else
            cleanupHappened();
        return result;
    }

    unsigned computeSize() const
    {
        const_cast<WeakListHashSet&>(*this).removeNullReferences();
        return m_set.size();
    }

    void checkConsistency() { } // To be implemented.

    bool removeNullReferences()
    {
        bool didRemove = m_set.removeIf([] (auto& value) { return !value.get(); });
        cleanupHappened();
        return didRemove;
    }

    T& first()
    {
        auto it = begin();
        ASSERT(it != end());
        return *it;
    }

    const T& first() const
    {
        auto it = begin();
        ASSERT(it != end());
        return *it;
    }

    T& takeFirst()
    {
        auto it = begin();
        auto& first = *it;
        remove(it);
        return first;
    }

    T* tryTakeFirst()
    {
        auto it = begin();
        if (it == end())
            return nullptr;
        auto* first = it.get();
        remove(it);
        return first;
    }

    T& last()
    {
        auto it = end();
        --it;
        return *it;
    }

    const T& last() const
    {
        auto it = end();
        --it;
        return *it;
    }

private:
    ALWAYS_INLINE void cleanupHappened() const
    {
        m_operationCountSinceLastCleanup = 0;
        m_maxOperationCountWithoutCleanup = std::min(std::numeric_limits<unsigned>::max() / 2, m_set.size()) * 2;
    }

    ALWAYS_INLINE unsigned increaseOperationCountSinceLastCleanup(unsigned count = 1) const
    {
        unsigned currentCount = m_operationCountSinceLastCleanup += count;
        return currentCount;
    }

    ALWAYS_INLINE void amortizedCleanupIfNeeded(unsigned count = 1) const
    {
        unsigned currentCount = increaseOperationCountSinceLastCleanup(count);
        if (currentCount > m_maxOperationCountWithoutCleanup)
            const_cast<WeakListHashSet&>(*this).removeNullReferences();
    }

    SetType m_set;
    mutable unsigned m_operationCountSinceLastCleanup { 0 };
    mutable unsigned m_maxOperationCountWithoutCleanup { 0 };
};

template<typename T, typename WeakPtrImplType>
size_t containerSize(const WeakListHashSet<T, WeakPtrImplType>& container) { return container.computeSize(); }

template<typename T, typename WeakPtrImplType>
inline auto copyToVector(const WeakListHashSet<T, WeakPtrImplType>& collection) -> Vector<typename WeakListHashSet<T, WeakPtrImplType>::KeyType> {
    return WTF::map(collection, [](auto& v) -> typename WeakListHashSet<T, WeakPtrImplType>::KeyType {
        return typename WeakListHashSet<T, WeakPtrImplType>::KeyType { v };
    });
}

}

using WTF::WeakListHashSet;

==> webkit2gtk_2.51.4.orig.tar.xz: webkitgtk-2.51.4/Source/WTF/wtf/WeakListHashSet.h <==
/*
 * Copyright (C) 2023 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <algorithm>
#include <wtf/ListHashSet.h>
#include <wtf/WeakPtr.h>

namespace WTF {

// Value will be deleted lazily upon rehash or amortized over time. For manual cleanup, call removeNullReferences().
template <typename T, typename WeakPtrImplType>
class WeakListHashSet final {
    WTF_DEPRECATED_MAKE_FAST_ALLOCATED(WeakListHashSet);
public:
    using KeyType = WeakPtr<T, WeakPtrImplType>;
    using SetType = ListHashSet<KeyType>;
    using AddResult = typename SetType::AddResult;

    template <typename ListHashSetType, typename IteratorType>
    class WeakListHashSetIteratorBase {
    public:
        using iterator_category = std::forward_iterator_tag;
        using value_type = T;
        using difference_type = ptrdiff_t;
        using pointer = T*;
        using reference = T&;

    protected:
        WeakListHashSetIteratorBase(ListHashSetType& set, IteratorType position)
            : m_set { set }
            , m_position { position }
            , m_beginPosition { set.m_set.begin() }
            , m_endPosition { set.m_set.end() }
        {
        }

        WeakListHashSetIteratorBase(const ListHashSetType& set, IteratorType position)
            : m_set { set }
            , m_position { position }
            , m_beginPosition { set.m_set.begin() }
            , m_endPosition { set.m_set.end() }
        {
        }

        ~WeakListHashSetIteratorBase() = default;

        T* get() { return m_position->get(); }
        const T* get() const { return get(); }
        T& operator*() { return *get(); }
        const T& operator*() const { return *get(); }
        T* operator->() { return get(); }
        const T* operator->() const { return get(); }

        void advance()
        {
            ASSERT(m_position != m_endPosition);
            ++m_position;
            m_set.increaseOperationCountSinceLastCleanup();
        }

        void advanceBackwards()
        {
            ASSERT(m_position != m_beginPosition);
            --m_position;
            while (m_position != m_beginPosition && !get())
                --m_position;
            m_set.increaseOperationCountSinceLastCleanup();
        }

        const ListHashSetType& m_set;
        IteratorType m_position;
        IteratorType m_beginPosition;
        IteratorType m_endPosition;
    };

    class WeakListHashSetIterator : public WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::iterator> {
    public:
        using Base = WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::iterator>;

        bool operator==(const WeakListHashSetIterator& other) const { return Base::m_position == other.Base::m_position; }

        using Base::get;
        using Base::operator*;
        using Base::operator->;

        WeakListHashSetIterator& operator++()
        {
            Base::advance();
            return *this;
        }

        WeakListHashSetIterator& operator--()
        {
            Base::advanceBackwards();
            return *this;
        }

    private:
        WeakListHashSetIterator(WeakListHashSet& map, typename SetType::iterator position)
            : Base { map, position }
        { }

        template <typename, typename> friend class WeakListHashSet;
    };

    class WeakListHashSetConstIterator : public WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::const_iterator> {
    public:
        using Base = WeakListHashSetIteratorBase<WeakListHashSet, typename SetType::const_iterator>;

        bool operator==(const WeakListHashSetConstIterator& other) const { return Base::m_position == other.Base::m_position; }

        using Base::get;
        using Base::operator*;
        using Base::operator->;

        WeakListHashSetConstIterator& operator++()
        {
            Base::advance();
            return *this;
        }

        WeakListHashSetConstIterator& operator--()
        {
            Base::advanceBackwards();
            return *this;
        }

    private:
        WeakListHashSetConstIterator(const WeakListHashSet& map, typename SetType::const_iterator position)
            : Base { map, position }
        { }

        template <typename, typename> friend class WeakListHashSet;
    };

    using iterator = WeakListHashSetIterator;
    using const_iterator = WeakListHashSetConstIterator;

    iterator begin() { return WeakListHashSetIterator(*this, m_set.begin()); }
    iterator end() { return WeakListHashSetIterator(*this, m_set.end()); }
    const_iterator begin() const { return WeakListHashSetConstIterator(*this, m_set.begin()); }
    const_iterator end() const { return WeakListHashSetConstIterator(*this, m_set.end()); }

    iterator find(const T* value)
    {
        increaseOperationCountSinceLastCleanup();
        return WeakListHashSetIterator(*this, m_set.find(value));
    }

    iterator find(const T& value)
    {
        return find(&value);
    }

    const_iterator find(const T* value) const
    {
        increaseOperationCountSinceLastCleanup();
        return WeakListHashSetConstIterator(*this, m_set.find(value));
    }

    const_iterator find(const T& value) const
    {
        return find(&value);
    }

    bool contains(const T* value) const
    {
        increaseOperationCountSinceLastCleanup();
        return m_set.contains(value);
    }

    bool contains(const T& value) const
    {
        return contains(&value);
    }

    AddResult add(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.add(value);
    }

    AddResult add(const T& value)
    {
        return add(&value);
    }

    AddResult appendOrMoveToLast(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.appendOrMoveToLast(value);
    }

    AddResult appendOrMoveToLast(const T& value)
    {
        return appendOrMoveToLast(&value);
    }

    bool moveToLastIfPresent(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.moveToLastIfPresent(value);
    }

    bool moveToLastIfPresent(const T& value)
    {
        return moveToLastIfPresent(&value);
    }

    AddResult prependOrMoveToFirst(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.prependOrMoveToFirst(value);
    }

    AddResult prependOrMoveToFirst(const T& value)
    {
        return prependOrMoveToFirst(&value);
    }

    AddResult insertBefore(const T* beforeValue, const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.insertBefore(beforeValue, value);
    }

    AddResult insertBefore(const T& beforeValue, const T& value)
    {
        return insertBefore(&beforeValue, &value);
    }

    AddResult insertBefore(iterator it, const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.insertBefore(it.m_position, value);
    }

    AddResult insertBefore(iterator it, const T& value)
    {
        return insertBefore(it, &value);
    }

    bool remove(const T* value)
    {
        amortizedCleanupIfNeeded();
        return m_set.remove(value);
    }

    bool remove(const T& value)
    {
        return remove(&value);
    }

    bool remove(iterator it)
    {
        if (it == end())
            return false;

        // WeakPtr<T> can become null while 'it' is live.
        if (!it.get())
            return false;

        auto result = m_set.remove(it.m_position);
        amortizedCleanupIfNeeded();
        return result;
    }

    void clear()
    {
        m_set.clear();
        cleanupHappened();
    }

    unsigned capacity() const { return m_set.capacity(); }

    bool isEmptyIgnoringNullReferences() const
    {
        if (m_set.isEmpty())
            return true;

        auto onlyContainsNullReferences = begin() == end();
        if (onlyContainsNullReferences) [[unlikely]]
            const_cast<WeakListHashSet&>(*this).clear();
        return onlyContainsNullReferences;
    }

    bool hasNullReferences() const
    {
        unsigned count = 0;
        for (auto _ : m_set) {
            UNUSED_VARIABLE(_);
            ++count;
        }

        bool result = count != m_set.size();
        if (result)
            increaseOperationCountSinceLastCleanup(count);
        else
            cleanupHappened();
        return result;
    }

    unsigned computeSize() const
    {
        const_cast<WeakListHashSet&>(*this).removeNullReferences();
        return m_set.size();
    }

    void checkConsistency() { } // To be implemented.

    void removeNullReferences()
    {
        m_set.removeWeakNullEntries();
        cleanupHappened();
    }

    T& first()
    {
        auto it = begin();
        ASSERT(it != end());
        return *it;
    }

    const T& first() const
    {
        auto it = begin();
        ASSERT(it != end());
        return *it;
    }

    T& takeFirst()
    {
        auto it = begin();
        auto& first = *it;
        remove(it);
        return first;
    }

    T* tryTakeFirst()
    {
        auto it = begin();
        if (it == end())
            return nullptr;
        auto* first = it.get();
        remove(it);
        return first;
    }

    T& last()
    {
        auto it = end();
        --it;
        return *it;
    }

    const T& last() const
    {
        auto it = end();
        --it;
        return *it;
    }

private:
    ALWAYS_INLINE void cleanupHappened() const
    {
        m_operationCountSinceLastCleanup = 0;
        m_maxOperationCountWithoutCleanup = std::min(std::numeric_limits<unsigned>::max() / 2, m_set.size()) * 2;
    }

    ALWAYS_INLINE unsigned increaseOperationCountSinceLastCleanup(unsigned count = 1) const
    {
        unsigned currentCount = m_operationCountSinceLastCleanup += count;
        return currentCount;
    }

    ALWAYS_INLINE void amortizedCleanupIfNeeded(unsigned count = 1) const
    {
        unsigned currentCount = increaseOperationCountSinceLastCleanup(count);
        if (currentCount > m_maxOperationCountWithoutCleanup)
            const_cast<WeakListHashSet&>(*this).removeNullReferences();
    }

    SetType m_set;
    mutable unsigned m_operationCountSinceLastCleanup { 0 };
    mutable unsigned m_maxOperationCountWithoutCleanup { 0 };
};

template<typename T, typename WeakPtrImplType>
size_t containerSize(const WeakListHashSet<T, WeakPtrImplType>& container) { return container.computeSize(); }

template<typename T, typename WeakPtrImplType>
inline auto copyToVector(const WeakListHashSet<T, WeakPtrImplType>& collection) -> Vector<typename WeakListHashSet<T, WeakPtrImplType>::KeyType> {
    return WTF::map(collection, [](auto& v) -> typename WeakListHashSet<T, WeakPtrImplType>::KeyType {
        return typename WeakListHashSet<T, WeakPtrImplType>::KeyType { v };
    });
}

}

using WTF::WeakListHashSet;
